

#include "first.h"
#include "std_opt.h"
#include "raster_util.h"
#include "clut.h"
#include <assert.h>
#include "rast_i.h"
#include "sbuf.h"
#include "sbuf_i.h"

using namespace raster;


//
// note: u's overflow after some number.. gotta fix this.
//

//
// size of blocks for bilinear interpolation
// 4 means 16,16 blocks higher values are faster but less perspectively 
// correct
//
const int y_block_bits = 4;
const int x_block_bits = 4;
const int x_block_dim = (1<<x_block_bits);
const int y_block_dim = (1<<y_block_bits);
const int max_x_blocks =  max_x_res/x_block_dim + 1;
const int max_y_blocks =  max_y_res/y_block_dim + 1;

extern "C" const void * cur_solid_clut;


const float shade_eps = 1.0f/256.0f;
const float p_min_shade   = shade_eps;

#define FSHL(exp)  ( (127+( exp )) << 23)


const int shade_fx_bits = 16;

const int uv_fx_bits = 20;

#define __uv_add (x_block_bits)

const float p_u_magic       = fshl_hf( 23 - uv_fx_bits +(__uv_add)  );


#define __shade_add (0)

//const float plt_shade_magic  =  fshl( 23 - shade_fx_bits +(x_block_bits)  );
const float plt_shade_magic  =  fshl( 23 - shade_fx_bits+__shade_add   );



// 	mov dl,127 + 23 - 20 + 4 + 8;
//float tp_v_magic = fshl_hf( 23 - uv_fx_bits + 4 + 8  - 5 );

// depends on size of texture
float p_v_magic;

const int i_p_min_shade = FSHL(23-shade_fx_bits + (__shade_add)) + (256*1  >>(__shade_add));
const int i_p_max_shade = FSHL(23-shade_fx_bits + (__shade_add)) + (256*255>>(__shade_add));


static u8 x_extents[max_y_blocks][2];


//
//  data stored at each point in grid
//


struct pt_dat
{
 int shade;
 int u, v;
 inline void operator = (const pt_dat & src) { shade = src.shade;u = src.u;v = src.v;}
};

//
// pt dats at last y
//
static pt_dat       last_y_dats[max_x_blocks];



// from a solid

//
// packed data created from pt_dats, and interpolated along y 
//
struct block_data
{
 __int64 uvs,d_dy_uvs;
};

block_data block_datas[max_x_blocks*2];


//
// 16 bit versions
//
extern "C" void __fastcall draw_span_block_init(const raster::texture * tmap);
extern "C" int  __fastcall draw_span_block(int y);
extern "C" int  __fastcall draw_span_block_solid(int y);



//
// number if fixed point bits to use for u and v
//


inline int get_x_block0( int x0 ) { return x0 >> x_block_bits; }
inline int get_x_block1( int x1 ) { return (x1+x_block_dim-1) >> x_block_bits; }



//
// pack u, v, shade into 1 64 bit number  
// u: 32 - v_bits
// v: v_bits + 16
// shade: 16

inline __int64 pack_uvshade(int u, int v, int shade)
{       // make 64 bit version
   __int64 rv;    
    ((int*)&rv)[1]  =  ((u<< (32 -uv_fx_bits) )) | ( (v>> (uv_fx_bits-8) )& 0xfff );
    (int&)rv        =  (v<< (40 -uv_fx_bits) ) | ( (shade ) &0xfffff);
   return rv;
} 




inline int fdif_y( int end, int start) { return (end-start  )>>y_block_bits; }
//inline int fdif_y( int end, int start) { return (end>>y_block_bits)-(start>>y_block_bits); }
   


#define SET_LINE_DAT(X0,X1)                        \
{                                                  \
   x_extents[yb][0] = u8(get_x_block0( X0 ));                        \
   x_extents[yb][1] = u8(get_x_block1( X1 ));                        \
}

const short double_pres = 0x200;
const short single_pres = 0x0;

// uncomment this for exceptions on divide by 0
//const short default_mask    = 0x07b;

const short default_mask    = 0x07f;
const short zero_divide_mask= 0x04;




block_data * start_block;
block_data * end_block;

extern "C" u32 * cur_end_span;
extern "C" u32 * cur_span;





static void make_block_data_1( int y, int x0, int x1, const rend_data_tex & cur_rd)
{
                
 pt_dat       p1; // current point that we are making
 

 const float dz_dx = cur_rd.dz[0];
 float zb_ = cur_rd.z0 + cur_rd.dz[1]*y + x0*dz_dx;

 const float  du_dx = cur_rd.du[0];
 float ub_ = cur_rd.u0 + cur_rd.du[1]*y + x0*du_dx;
 
 const float  dv_dx = cur_rd.dv[0];
 float vb_ = cur_rd.v0 + cur_rd.dv[1]*y + x0*dv_dx;

 const float  ds_dx = cur_rd.dsh[0];
 float sb_ = cur_rd.sh0 + cur_rd.dsh[1]*y + x0*ds_dx;

 float z = 1.0f/zb_;

 do
 {

     float u = z * ub_;
     float v = z * vb_;
     float s = z * sb_;

     zb_ += dz_dx;
     ub_ += du_dx;
     vb_ += dv_dx;
     sb_ += ds_dx;
     
     (float&)p1.u = u + p_u_magic;
     (float&)p1.v = v + p_v_magic;
  
   
     (float&)p1.shade = (s * (1.0 -  z * raster::haze_far_z_inv  )) + plt_shade_magic;

	 p1.shade &=  ~( (1<<x_block_bits) - 1 );

     test_min( p1.shade,  i_p_max_shade); 
     test_max( p1.shade,  i_p_min_shade); 

     
	// p1.u = 0;
    
 //   gen_light_data * lg = lighting_lights;
 //   __comp_shade( p1.shade) 
   


     // make next z
    
     z = 1.0f/zb_;

   // is this the first block?
     pt_dat & p0 = last_y_dats[x0];

     // make new value
     // are we could be going back too far?

      block_data & bd = block_datas[x0];
    
      bd.uvs      = pack_uvshade( p0.u,p0.v,p0.shade);
    
      bd.d_dy_uvs = pack_uvshade( fdif_y(p1.u,p0.u),
                                  fdif_y(p1.v,p0.v),
                                  fdif_y(p1.shade,p0.shade)  );
         // shift down by 1
      p0 = p1;


  } while( ++x0, x0 <= x1 );
}



//
// if highest bit of light_flags is set, this is drawn 
// as a flat shaded poly whose intensity is == light_flags
//
void __fastcall sb_fill_plt(rend_data * _rd, u32 * span_start,u32 * end_span)
{
 rend_data_tex & cur_rd = *(rend_data_tex *)_rd;

  draw_span_block_init(cur_rd.tmap );

  scale_float( cur_rd.dz[0], x_block_bits  );
  scale_float( cur_rd.du[0], x_block_bits  );
  scale_float( cur_rd.dv[0], x_block_bits  );
  scale_float( cur_rd.dsh[0], x_block_bits  );


 int y_start =  get_y( *span_start);


  set_single_pres();
  
  cur_end_span = end_span;


  {
   // calculate x extents
   u32 * span = span_start;
   int cur_y =  y_start;
   int yb = 0;
 
   u32 l_min_x = 1024, l_max_x=0;
   for(;;)
   {
    u32 min_x = get_x0( *span );
    u32 max_x = get_x1( *span );
    for(;;)
    {
       ++span;
       if( span == end_span)
       {
        SET_LINE_DAT( __min(min_x,l_min_x),__max(max_x,l_max_x) );
        ++yb;
        SET_LINE_DAT( min_x,max_x );
        goto SPANS_DONE;
       }
       if( ((int)span &63) == 0 ) 
          span = (u32*)span[-16];

       int y = get_y(*span);
       if( y >= cur_y + y_block_dim)
         break;
    
       u32 x0 = get_x0( *span );
       u32 x1 = get_x1( *span );

   
       test_min(min_x, x0);
       test_max(max_x, x1);
    }
    
    SET_LINE_DAT( __min(min_x,l_min_x),__max(max_x,l_max_x) );

    l_min_x = min_x;
    l_max_x = max_x;

    cur_y+= y_block_dim;
     ++yb;
   }
   SPANS_DONE:;
 
 }

    //u32 * span = span_start;
    cur_span = span_start;
  
    // make first block
    int yb = 0;
    make_block_data_1(y_start,x_extents[0][0], x_extents[0][1] , cur_rd);

    u32 cur_y = y_start;
    for(;;++yb)
    {
     // make next line
    
      u32 eb0 =  x_extents[yb+1][0];
      u32 eb1 =  x_extents[yb+1][1];
    
      u32 end_y = cur_y + y_block_dim;
      make_block_data_1(end_y,eb0,eb1 , cur_rd);

     // make_packed_data(eb0,eb1);
      start_block = &block_datas[eb0];
      end_block   = &block_datas[eb1];

      for(;;)
      {
       int finished = draw_span_block(cur_y);
       if( finished )
         goto FINISHED;

     
       ++cur_y;
       if( cur_y >= end_y)
         break;
      
      } // for ys

   } // for each y block
  
 
FINISHED:  ;
  set_double_pres();

  return;

};



 poly *  raster::submit_plt(const vert_3s & v0,
                          const vert_3s & v1,
                          const vert_3s & v2,
                          const  texture* tmap,
                          const float * uv0, const float * uv1, const float * uv2)
 {
	STD_P_SETUP()

 
  push_poly(plt_polys,nply );

  rend_data_tex * rd = alloc_rd(rd);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  SB_LB_TO_SCREEN_Z()

  SB_LB_TO_SCREEN_P(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade );
  SB_LB_TO_SCREEN_P(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN_P(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
  
 // rd->u0    *= 2.0f;
 // rd->du[0] *= 2.0f;
 // rd->du[1] *= 2.0f;


  return nply;
}



 poly *  raster::submit_pft(const vert_3 & v0,
                           const vert_3 & v1,
                           const vert_3 & v2,
                           float shade,
                           const  texture* tmap,
                           const float * uv0, const float * uv1, const float * uv2)
 {
  STD_P_SETUP()

 
  push_poly(plt_polys,nply );

  rend_data_tex * rd = alloc_rd(rd);
  nply->reset(rd);
 
  // set texture
  rd->tmap = tmap;


  SB_LB_TO_SCREEN_Z()

  SB_LB_TO_SCREEN_P(rd->sh0,rd->dsh, shade, shade, shade );
  SB_LB_TO_SCREEN_P(rd->u0,rd->du,uv0[0],uv1[0],uv2[0]);
  SB_LB_TO_SCREEN_P(rd->v0,rd->dv,uv0[1],uv1[1],uv2[1]);
  

  return nply;
}




static void make_block_data_solid( int y, int x0, int x1, const rend_data_gour & cur_rd)
{
                
 pt_dat       p1; // current point that we are making
 

 const float dz_dx = cur_rd.dz[0];
 float zb_ = cur_rd.z0 + cur_rd.dz[1]*y + x0*dz_dx;


 const float  ds_dx = cur_rd.dsh[0];
 float sb_ = cur_rd.sh0 + cur_rd.dsh[1]*y + x0*ds_dx;

 float z = 1.0f/zb_;

 do
 {

     float s = z * sb_;

     zb_ += dz_dx;
     sb_ += ds_dx;
     
   
     (float&)p1.shade = (s * (1.0 -  z * raster::haze_far_z_inv  )) + plt_shade_magic;

     test_min( p1.shade,  i_p_max_shade); 
     test_max( p1.shade,  i_p_min_shade); 


     // make next z
    
     z = 1.0f/zb_;

   // is this the first block?
     pt_dat & p0 = last_y_dats[x0];

     // make new value
     // are we could be going back too far?

      block_data & bd = block_datas[x0];
    
      bd.uvs      = pack_uvshade( 0,0,p0.shade);
    
      bd.d_dy_uvs = pack_uvshade( fdif_y(0,0),
                                  fdif_y(0,0),
                                  fdif_y(p1.shade,p0.shade)  );
         // shift down by 1
      p0 = p1;


  } while( ++x0, x0 <= x1 );
}

 

//
// if highest bit of light_flags is set, this is drawn 
// as a flat shaded poly whose intensity is == light_flags
//
void __fastcall sb_fill_pls(rend_data * _rd, u32 * span_start,u32 * end_span)
{
  rend_data_gour & cur_rd = *(rend_data_gour *)_rd;
  cur_solid_clut = cur_rd.clut;

//  draw_span_block_init(cur_rd.tmap );

  scale_float( cur_rd.dz[0], x_block_bits  );
  scale_float( cur_rd.dsh[0], x_block_bits  );


  int y_start =  get_y( *span_start);


  set_single_pres();
  
  cur_end_span = end_span;


  {
   // calculate x extents
   u32 * span = span_start;
   int cur_y =  y_start;
   int yb = 0;
 
   u32 l_min_x = 1024, l_max_x=0;
   for(;;)
   {
    u32 min_x = get_x0( *span );
    u32 max_x = get_x1( *span );
    for(;;)
    {
       ++span;
       if( span == end_span)
       {
        SET_LINE_DAT( __min(min_x,l_min_x),__max(max_x,l_max_x) );
        ++yb;
        SET_LINE_DAT( min_x,max_x );
        goto SPANS_DONE;
       }
       if( ((int)span &63) == 0 ) 
          span = (u32*)span[-16];

       int y = get_y(*span);
       if( y >= cur_y + y_block_dim)
         break;
    
       u32 x0 = get_x0( *span );
       u32 x1 = get_x1( *span );

   
       test_min(min_x, x0);
       test_max(max_x, x1);
    }
    
    SET_LINE_DAT( __min(min_x,l_min_x),__max(max_x,l_max_x) );

    l_min_x = min_x;
    l_max_x = max_x;

    cur_y+= y_block_dim;
     ++yb;
   }
   SPANS_DONE:;
 
 }

    //u32 * span = span_start;
    cur_span = span_start;
  
    // make first block
    int yb = 0;
    make_block_data_solid(y_start,x_extents[0][0], x_extents[0][1] , cur_rd);

    u32 cur_y = y_start;
    for(;;++yb)
    {
     // make next line
    
      u32 eb0 =  x_extents[yb+1][0];
      u32 eb1 =  x_extents[yb+1][1];
    
      u32 end_y = cur_y + y_block_dim;
      make_block_data_solid(end_y,eb0,eb1 , cur_rd);

     // make_packed_data(eb0,eb1);
      start_block = &block_datas[eb0];
      end_block   = &block_datas[eb1];

      for(;;)
      {
       int finished = draw_span_block_solid(cur_y);
       if( finished )
         goto FINISHED;

     
       ++cur_y;
       if( cur_y >= end_y)
         break;
      
      } // for ys

   } // for each y block
  
 
FINISHED:  ;
  set_double_pres();

  return;

};


poly *  raster::submit_pls(const vert_3s & v0,
                          const vert_3s & v1,
                          const vert_3s & v2,
                          const  solid_surface_desc * sd
                          )
{
  STD_P_SETUP()

 
  push_poly(pls_polys,nply );

  rend_data_gour * rd = alloc_rd(rd);
  nply->reset(rd);
 
  // set texture
  rd->clut = sd->clut;


  SB_LB_TO_SCREEN_Z()

  SB_LB_TO_SCREEN_P(rd->sh0,rd->dsh, v0.shade, v1.shade, v2.shade );
  

  return nply;
}



